from unittest import TestCase


class Test(TestCase):
    def test__parse_cols(self):
        from src.pd_helper.tests.test__config import BigDF
        df = BigDF(num_cols=1000, num_rows=10000, num_size=1000).a_dataframe()
        from src.pd_helper.utils._parse_cols import _parse_cols
        df = _parse_cols(df)
        cols = list(df.columns)
        has_space = False

        if any(' ' in i for i in cols):
            has_space = True

        self.assertFalse(has_space)



import logging
import sys

def _configuration(echo=False):
    """Runs configuration for logging.
    """
    if echo:
        level = logging.INFO
    else:
        level = None

    logger = logging.getLogger("pipeline").getChild("configuration")
    format = "%(asctime)s:%(name)s:%(levelname)s:%(message)s"

    log_formatter = logging.Formatter(format)
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(log_formatter)

    logging.basicConfig(format=format, level=level)
    logging.getLogger().addHandler(stream_handler)

    if echo:
        logger.info("Logging configured.")






import pandas as pd
import numpy as np
import numpy.lib.recfunctions as rfn

import os
import uuid
import shortuuid
import random


class MakeData:
    """
    MakeData is a class of functions that returns a perfectly imperfect dataframe of various dtypes for testing.
    """
    def __init__(self):
        self.hourly = np.timedelta64(1, 'h')
        self.daily = np.timedelta64(1, 'D')
        self.today = pd.Timestamp.now()

    def make_random_dates(self, start, end, n):
        """

        :param start: A string date like '2010-01-01', should be less than end date
        :param end: A string date like '2020-01-01', should be greater than start date
        :param n: number of records to return
        :return: a series of random datetime stamps in a given range

        works cited: https://stackoverflow.com/questions/50559078/generating-random-dates-within-a-given-range-in-pandas
        """
        start = pd.to_datetime(start)
        end = pd.to_datetime(end)
        start_u = start.value // 10 ** 9
        end_u = end.value // 10 ** 9
        return pd.DatetimeIndex((10 ** 9 * np.random.randint(start_u, end_u, n, dtype=np.int64)).view('M8[ns]'))

    def make_df(self, make_default=True, perfect=False, size=100000, col_names=None, values=None):
        """
        :param make_default: make a Pandas Dataframe of imperfect data types
        :param perfect: no purposeful mistakes in data
        :param size: number of rows in dataframe
        :param col_names: optional col names, or default
        :param values: optional col values, or default
        :return: a dataframe of randomly generated data
        """
        df_out = pd.DataFrame()

        if make_default:

            columns = ['Object ID', 'Item Name', 'Retrieved Date', 'Retrieved', 'Condition', 'Sector', 'Status',
                       'Status Date', 'Weight']
            n_cols = len(columns)
            n_sectors = 8
            sectors = [x for x in range(n_sectors)]

            if perfect:
                conditions = ['Excellent', 'Poor', 'Good', 'Spare Parts', 'Trash']
                items = ['Lighter', 'Toaster', 'YoYo']
            else:
                conditions = ['Excellent', 'excellent', 'excelent', 'Poor', 'poor', 'Good', 'good', 'Spare Parts',
                              'Trash', 'trsh']
                items = ['lighter', 'Lighter', 'toster', 'Toaster', 'YoYo', 'yo-yo']

            statuses = ['Inventoried', 'Inventoried', 'Repaired', 'Repaired', 'Pending Repair', 'Pending Repair',
                        'Pending Repair', 'Pending Repair', 'Pending Inventory', 'Pending Inventory',
                        'Pending Inventory', 'Missing']

            def filler():
                arr = np.zeros((size, n_cols))

                dtype = str(arr.dtype)
                dtypes = np.dtype([(n, dtype) for n in columns])

                structured = rfn.unstructured_to_structured(arr, dtypes)
                df = pd.DataFrame.from_records(data=structured)

                df['Object ID'] = df.apply(lambda x: shortuuid.uuid(), axis=1)

                df['Retrieved Date'] = self.make_random_dates('2010-01-01', '2020-12-31', size)

                # https://stackoverflow.com/questions/6824681/get-a-random-boolean-in-python
                df['Retrieved'] = df.apply(lambda x: bool(random.getrandbits(1)), axis=1)

                df['Sector'] = df.apply(lambda x: np.random.choice(sectors, 1, replace=True)[0], axis=1)

                df['Condition'] = df.apply(lambda x: np.random.choice(conditions, 1, replace=True)[0], axis=1)

                df['Status'] = df.apply(lambda x: np.random.choice(statuses, 1, replace=True)[0], axis=1)

                df['Status Date'] = self.make_random_dates('2020-01-01', '2020-12-31', size)

                df['Item Name'] = df.apply(lambda x: np.random.choice(items, 1, replace=True)[0], axis=1)

                def lighters():
                    return np.random.uniform(low=1, high=11)

                def toasters():
                    return np.random.uniform(low=20, high=48)

                def yoyos():
                    return np.random.uniform(low=5, high=16)

                cond1 = ['lighter', 'Lighter']
                cond2 = ['Toaster', 'toster']
                cond3 = ['YoYo', 'yo-yo']

                df['Weight'] = df.apply(lambda x: lighters(), axis=1)

                df['Weight'] = np.where(df['Item Name'].isin(cond1), df.apply(lambda x: lighters(), axis=1),
                                        np.where(df['Item Name'].isin(cond2), df.apply(lambda x: toasters(), axis=1),
                                                 np.where(df['Item Name'].isin(cond3),
                                                          df.apply(lambda x: yoyos(), axis=1), 0.0)))

                best = ['Excellent', 'excellent', 'excelent', 'Good', 'good']
                mid = ['Good', 'good', 'Spare Parts']
                worst = ['Poor', 'poor', 'Trash', 'trsh']

                def rand_bool(percent=50):
                    # https://stackoverflow.com/questions/14324472/random-boolean-by-percentage
                    return random.randrange(100) < percent

                df['Retrieved'] = np.where(df['Condition'].isin(best),
                                           df.apply(lambda x: rand_bool(percent=70), axis=1),
                                           np.where(df['Condition'].isin(mid),
                                                    df.apply(lambda x: rand_bool(percent=40), axis=1),
                                                    np.where(df['Condition'].isin(worst),
                                                             df.apply(lambda x: rand_bool(percent=10), axis=1), False)))

                df['Status'] = np.where((df['Retrieved'] == False & df['Condition'].isin(worst)),
                                        'Compacted Trash Pile', df['Status'])
                df['Status'] = np.where((df['Retrieved'] == False & df['Condition'].isin(mid)),
                                        'Left in Sector', df['Status'])
                df['Status'] = np.where((df['Retrieved'] == False & df['Condition'].isin(best)),
                                        'Left in Sector', df['Status'])

                df = df.sort_values(by='Retrieved Date').reset_index(drop=True)

                def shuffle(yr):
                    full = (list(str(yr)))
                    inner = full[1:3]
                    inner.reverse()
                    full[1:3] = inner
                    yr = int(''.join(full))
                    return yr

                if not perfect:
                    n_samples = int(size * .002)

                    if n_samples < 1:
                        n_samples = 1

                    s1 = df.sample(n=n_samples, replace=False, random_state=0)
                    s1['Retrieved Date'] = s1['Retrieved Date'].astype('object')
                    s1['Retrieved Date'] = s1['Retrieved Date'].apply(lambda x: x.replace(year=shuffle(x.year), day=np.random.randint(1, 25)))
                    df['Retrieved Date'].update(s1['Retrieved Date'])

                    s2 = df.sample(n=n_samples, replace=False, random_state=21)
                    s2['Retrieved Date'] = s2['Retrieved Date'].apply(lambda x: '')
                    df['Retrieved Date'].update(s2['Retrieved Date'])

                    df['Retrieved Date'] = pd.to_datetime(df['Retrieved Date'])

                    s3 = df.sample(n=n_samples, replace=False, random_state=42)
                    s3['Retrieved'] = s3['Retrieved'].apply(lambda x: '')
                    df['Retrieved'].update(s3['Retrieved'])

                    n_samples = int(size * .1)

                    if n_samples < 1:
                        n_samples = 1

                    s4 = df.sample(n=n_samples, replace=False, random_state=42)
                    s4['Item Name'] = s4['Item Name'].apply(lambda x: '')
                    df['Item Name'].update(s4['Item Name'])

                return df

            df_out = filler()

        return df_out



def _mem_usage(df):
    """
    A helper function that evaluates memory in a DataFrame
    :param df: A Pandas DataFrame
    :return: An f-string that indicates DataFrame memory usage in MB.
    :source: https://gist.github.com/enamoria/fa9baa906f23d1636c002e7186516a7b
    """
    mem = df.memory_usage().sum() / 1024 ** 2

    return '{:.2f} MB'.format(mem)



from unittest import TestCase

import pandas as pd

class Test(TestCase):

    def test__reduce_precision(self):
        from src.pd_helper.maker import MakeData
        from src.pd_helper.utils._reduce_precision import _reduce_precision
        faker = MakeData()
        df = faker.make_df()
        df_cols = list(df.columns)

        # function params passed by parent
        date_strings = ['_date', 'date_']
        bool_types = [True, False, 'true', 'True', 'False', 'false']
        categorical_ratio = .1
        categorical_threshold = 20
        final_default_dtype = 'string'
        enable_mp = False

        s = _reduce_precision(df["Object ID"]
                              , date_strings=date_strings
                              , bool_types=bool_types
                              , categorical_ratio=categorical_ratio
                              , categorical_threshold=categorical_threshold
                              , final_default_dtype=final_default_dtype
                              , enable_mp=enable_mp
                              )
        data_type = str(s.dtype)
        assert data_type == 'string'

        s = _reduce_precision(df['Item Name']
                              , date_strings=date_strings
                              , bool_types=bool_types
                              , categorical_ratio=categorical_ratio
                              , categorical_threshold=categorical_threshold
                              , final_default_dtype=final_default_dtype
                              , enable_mp=enable_mp
                              )
        data_type = str(s.dtype)
        assert data_type == 'category'

        s = _reduce_precision(df["Retrieved Date"]
                              , date_strings=date_strings
                              , bool_types=bool_types
                              , categorical_ratio=categorical_ratio
                              , categorical_threshold=categorical_threshold
                              , final_default_dtype=final_default_dtype
                              , enable_mp=enable_mp
                              )

        data_type = str(s.dtype)
        assert data_type == 'datetime64[ns]'

        s = _reduce_precision(df['Retrieved']
                              , date_strings=date_strings
                              , bool_types=bool_types
                              , categorical_ratio=categorical_ratio
                              , categorical_threshold=categorical_threshold
                              , final_default_dtype=final_default_dtype
                              , enable_mp=enable_mp
                              )
        data_type = str(s.dtype)
        assert data_type == 'boolean'

        s = _reduce_precision(df["Sector"]
                              , date_strings=date_strings
                              , bool_types=bool_types
                              , categorical_ratio=categorical_ratio
                              , categorical_threshold=categorical_threshold
                              , final_default_dtype=final_default_dtype
                              , enable_mp=enable_mp
                              )
        data_type = str(s.dtype)
        assert data_type == 'int8'

        s = _reduce_precision(df["Weight"]
                              , date_strings=date_strings
                              , bool_types=bool_types
                              , categorical_ratio=categorical_ratio
                              , categorical_threshold=categorical_threshold
                              , final_default_dtype=final_default_dtype
                              , enable_mp=enable_mp
                              )
        data_type = str(s.dtype)
        assert data_type == 'float16'









import pandas as pd
import numpy as np


class BigDF:
    def __init__(self, num_cols, num_size, num_rows):
        self.num_rows = num_rows
        self.lst_of_cols = [f'Col {i}' for i in range(num_cols)]

        lst_of_floats = [np.random.uniform(0, 100) for i in range(num_cols//2)]
        lst_of_ints = list(np.random.randint(num_rows, size=num_size)) * (num_cols//2)

        self.lst_of_nums = lst_of_ints + lst_of_floats

        self.data = {k: v for (k, v) in zip(self.lst_of_cols, self.lst_of_nums)}

    def a_dataframe(self):
        df = pd.DataFrame(self.data, index=list(range(self.num_rows)))
        return df



import pandas as pd
import multiprocessing as mp
from tqdm import tqdm
from functools import partial
import logging


from pd_helper.utils._mem_usage import _mem_usage
from pd_helper.utils._parse_cols import _parse_cols
from pd_helper.utils._reduce_precision import _reduce_precision
from pd_helper.utils._configuration import _configuration

def optimize(df
             , parse_col_names=True
             , enable_mp=False
             , mp_processors=None
             , date_strings=None
             , exclude_cols=None
             , special_mappings=None
             , bool_types=None
             , categorical_ratio=.1
             , categorical_threshold=20
             , final_default_dtype='string'
             , echo=False
             ):
    """
    Optimize a Pandas DataFrame by applying least precision to column dtypes.

    :params
    ----------
    df: DataFrame, Required
        a pandas dataframe that needs optimization
    parse_col_names: bool, default True
        If passed, returns columns as lower case without spaces
    enable_mp: bool, default to False
        If passed, runs optimization on columns in parallel with multiprocessing
    mp_processors: integer, default to None
        If None, default to half of available processes from mp.cpu_count(), only effective if mp is enabled.
    date_strings: list of string, default to None
        If None, default to a list of strings that indicate date columns -> ['_date', 'date_']
    exclude_cols: list of string, default to None
        If None, default to a list of strings that indicate columns to exclude from optimization -> ['exclude_this_col', 'exclude_another_col'].
        Note, excluded columns are returned, just not run through the optimizer.
    special_mappings: dictionary of {string: list of strings}, default to None
        Optional. If provided indicate a special mapping for col types with a dictionary, these are excluded from auto optimization.
        Note, the key is the desired dtype and the value is a list of column names.
    bool_types: list of string, default to None
        If None, default to a list of bool and semantic string values that indicate there is a boolean dtype such
        as True False, etc. -> [True, False, 'true', 'True', 'False', 'false']
    categorical_ratio: float, default to None
        If None, default to .1 (10%). Evaluates the ratio of unique values in the column to determine if categorical dtype is worthwhile.
        Note, if cat ratio less than 10%, then, the column is assigned categorical.
    categorical_threshold: integer, default to None
        If None, default to 20. Evaluates the len() of unique values in a column to determine if categorical dtype is worthwhile.
        Note, if the number of unique values is less than 20, categorical dtype is assigned.
    final_default_dtype: string, default to None
        If None, default to "string" dtype. This is the catch all dtype to assign if optimizer does not assign one by rule.
    echo: bool, default to False.
        If True, echo progress, show progress, and display metrics.

    *Note when calling function with multiprocessing enabled, ensure that the function is called from a module such as:

        if __name__ == "__main__":
            df = pd.DataFrame(<Some data here>)

            df = optimize(df)

    """
    _configuration(echo=echo)

    start_mem = _mem_usage(df)

    logging.info(f'Starting DataFrame Optimization. Starting with {start_mem} memory.')
    if parse_col_names:
        logging.info('Converting Column Names to Lower Case and Removing Spaces')
        df = _parse_cols(df)
    cols_to_convert = []
    # a default of strings that indicate the column is some kind of datetime column
    if date_strings is None:
        date_strings = ['_date', 'date_']
    # make a list of strings from all available dataframe columns
    cols_to_convert = [i for i in df.columns]
    # accommodate any special user-defined mappings

    def f(s, c):
        """
        A Helper Function to assigning special mappings
        :param s: Pandas Series
        :param c: String
        :return: Pandas Series
        """
        logging.info(f'Applying special mappings for {c}')
        if c == 'category':
            s = pd.Categorical(s)
        elif c == 'datetime':
            s = pd.to_datetime(s, errors='coerce')
        else:
            s = s.astype(c)
        return s

    special_exclusions = []

    if special_mappings is not None:
        for k, v in special_mappings.items():
            for i in v:
                df[i] = f(df[i], k)
                special_exclusions.append(i)

    # exclude columns if a list is provided
    if exclude_cols is not None:
        cols_to_convert = [i for i in cols_to_convert if i not in exclude_cols]
    # by default, if special mappings are provided, exclude them from auto optimization
    if special_exclusions:
        cols_to_convert = [i for i in cols_to_convert if i not in special_exclusions]
    # by default, a list of values to be explicitly treated as bools
    if bool_types is None:
        bool_types = [True, False, 'true', 'True', 'False', 'false']
        # TODO: account for only T or only F or 1/0 situations
    if mp_processors is None:
        CPUs = mp.cpu_count() // 2
    # by default, enable multiprocessing to run optimizations in parallel
    if enable_mp:
        logging.info('Starting optimization process with multiprocessor.')
        # break out the dataframe into a list of series to be worked on in parallel
        lst_of_series = [df[d] for d in cols_to_convert]

        pool = mp.Pool(CPUs)

        if echo:
            # run with progress bar in terminal
            pbar = tqdm(lst_of_series, desc='Running DataFrame Optimization with multiprocessing')
        else:
            # use the same pbar object but iterate over a list with no progress bar
            pbar = lst_of_series

        _reduce_precision_ = partial(_reduce_precision
                                     , date_strings=date_strings
                                     , bool_types=bool_types
                                     , categorical_ratio=categorical_ratio
                                     , categorical_threshold=categorical_threshold
                                     , final_default_dtype=final_default_dtype
                                     , enable_mp=enable_mp
                                     )
        list_of_converted = list(pool.imap(_reduce_precision_, pbar))
        pool.close()
        pool.join()

        # update the dataframe based on the converted records
        for (col_name, col_series) in list_of_converted:
            df[col_name] = col_series
    else:
        logging.info('Starting optimization process in series.')
        logging.info('Optimization can be run in parallel with multiprocessing'
                     ', set enable_mp to True and see function params.')

        df[cols_to_convert] = df[cols_to_convert].apply(lambda x: _reduce_precision(x
                                                                                    , date_strings=date_strings
                                                                                    , bool_types=bool_types
                                                                                    , categorical_ratio=categorical_ratio
                                                                                    , categorical_threshold=categorical_threshold
                                                                                    , final_default_dtype=final_default_dtype
                                                                                    , enable_mp=enable_mp
                                                                                    ))

    logging.info(f'Converted DF with new dtypes as follows:\n{df.dtypes}')
    end_mem = _mem_usage(df)
    logging.info(f'Completed DataFrame Optimization. Ending with {end_mem} memory.')

    return df








